/**********************************************************************
 playmp3 - Copyright (C) 1997 - Hans Petter Bieker <zerium@pe049.persbraten.vgs.no>
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#include <signal.h>
#include <getopt.h>
#include <pthread.h>
#include <linux/sched.h>
#include <linux/unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <semaphore.h>
#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/stat.h>
#include <linux/soundcard.h>

#include "playmp3.h"
#include "mp3info.h"
#include "wav.h"

int writewait = DEFWRITEWAIT, maxconbuffers = DEFMAXCONBUFFERS, buffersize = DEFBUFFERSIZE;

int verbose = 0,quiet = 0;
pthread_t * read_thread;
char * binname;

struct prodcons {
  int readpos, writepos;        /* positions for reading and writing */
  sem_t sem_read;               /* number of elements available for reading */
  sem_t sem_write;              /* number of locations available for writing */
  int readloop;                 /* the current state of thre readloop */
  char ** buffer;               /* the actual data */
  wave_header * whdr;
};

struct prodcons buffer;

void parsecfgfile(FILE *File);
void print_mp3info(char * filename);
void option_req_argument (void);
void version(void);;
int playmp3(char * filename);
void init(void);
void deinit(void);
void player_crash(void);
void usage(int error);
void * readloop (char * filename);
void writeloop (char * filename);

int header_check(wave_header *head) {
	unsigned p;
	
	p = head->length - head->data_length;

	if ( (p != 36) && (p != 37) ) { 
		err(0,"Wrong length difference.");
		return 0;
	}

	if ( strncmp(head->main_chunk,"RIFF",4) ) {
		err(0,"Not in RIFF format.");
		return 0;
	}

	if ( strncmp(head->chunk_type,"WAVE",4) ) {
		err(0,"Not a WAVE type.");
		return 0;
	}
		
	if ( strncmp(head->sub_chunk,"fmt ",4) ) {
		err(0,"Not a WAVE format.");
		return 0;
	}

	if ( (head->sample_fq * head->byte_p_spl) != head->byte_p_sec ) 
		err(0,"Incorrect sampling rate.");

	return 1;
}


void print_mp3info(char * filename)
{
  int  file;
  mp3file mp3;
  tune song;
  char format[FORMAT_LEN+1];
  int  errorno=0;
                                   
  memset(&song,0,sizeof(song));
  song.gennum=-1;
  strcpy(format,"Playing %f:\n%IT%It%t%!a, %IE%IE%Ia(%a), %IE%IE%b kbit/s %M %ICCRC %IE%(%mm %ss%)");
  if ((file=mp3open(&mp3,filename,0))==-1) {
    fprintf(stderr,"Error: Could not open file %s.\n",filename);
    errorno=1;
  } else if (!(mp3.hashdr=gethdr(file,&mp3.hdr))) {
    fprintf(stderr,"Error: no SYNC in file %s.\n",filename);
    errorno=2;
  } else {
    mp3calc(&mp3);
    
    memset(&mp3.tag,0,sizeof(mp3.tag));
    mp3.hastag=gettag(file,&mp3.tag); 
    printstd(&mp3,format,filename,NULL);
  }
  close(file); 
  if (errorno) exit (errorno);
}

void option_req_argument(void)
{
  fprintf(stderr,"%s: option requires an argument -- \n", binname);
  exit (1);
}

void player_crash(void)
{
  fprintf(stderr,"\n%s: An error accour in player\n",binname);
  exit (2);
}

void usage(int error)
{
  fprintf(error?stderr:stdout,"usage: %s [ -options ] file1 [file2 ...]
  -c, --maxconbuffers x   max number of concurrent buffers (should be >0)
  -s, --buffersize x      buffersize in kB (shuld be >0)
  -w, --write-wait x      seconds to wait before playing
  -q, --quiet             quiet mode; don't display any info about the sample
  -h, --help              display this help and exit
  -v, --version           output version information and exit
      --verbose           explain what is being done
",binname);
  exit(error?1:EXIT_SUCCESS);
}

void version(void)
{
  printf("%s: %s\n",binname,PROGVERSION);
  exit(EXIT_SUCCESS);
}

void * readloop (char * filename)
{
  FILE * instream;
  char * commandline;

  if (!(buffer.whdr  = malloc(sizeof(wave_header)))) {
    err(1,"Unable to allocate wave header");
    goto errxit;
  }

  if (!(commandline  = malloc(1024))) {
    err(1,"Unable to allocate commandline string");
    goto errxit;
  }
  /* get header */
  sprintf(commandline,"(l3dec -wav -fn 0 '%s' /tmp/head.in;mkmp3head /tmp/head.in /tmp/head.out&&cat /tmp/head.out&&rm -f /tmp/head.in /tmp/head.out) 2> /dev/null\n", filename);
  if (!(instream = popen(commandline, "r"))) {
    err(1,"Error starting l3dec #1");
    goto errxit;
  }
  if (fread(buffer.whdr,sizeof(wave_header), 1,instream) < 0) {
    err(1,"Error reading header");
    goto errxit;
  }
  pclose(instream);

  /* rest */
  sprintf(commandline,"l3dec -wav '%s' -sto 2>/dev/null", filename, filename);
  if (!(instream = popen(commandline, "r"))) {
    err(1,"Error starting l3dec #2");
    goto errxit;
  }
  free(commandline);

  buffer.readloop = ACTIVE_UNCOMPRESSING;
  while (!feof(instream)) {
    /* Wait until buffer is not full */
    sem_wait(&buffer.sem_write);
    /* Write the data and advance write pointer */
    if (verbose) {putchar('a');fflush(stdout);}
    if (!(buffer.buffer[buffer.writepos] = malloc(buffersize))) {
      err(1,"Unable to allocate audio uncompress buffer");
      goto errxit;
    }
    fread(buffer.buffer[buffer.writepos], buffersize, 1, instream);
    buffer.writepos++;
    if (buffer.writepos >= maxconbuffers) buffer.writepos = 0;
    /* Signal that the buffer contains one more element for reading */
    sem_post(&buffer.sem_read);
  }
  buffer.readloop = ACTIVE_FINISH;
  pclose(instream);
  return (void *)0;

errxit:
  pclose(instream);
  pthread_kill(pthread_self(),SIGPIPE);
}

void writeloop (char * filename)
{
  int tmp;
  int audio = -1;                       /* /dev/dsp audio file descriptor */
  int abuf_size;                        /* audio buffer size for I/O */
  char *iobuf;                          /* allocated I/O buffer */
  char tbuf[80];                        /* Temporary buffer */
  header_data data = {0,0,0};

  while (buffer.readloop == ACTIVE_INIT) sleep(1); /* wait for header - bad code.. */
  if ( !header_check(buffer.whdr) ) return; 

  data.speed = buffer.whdr->sample_fq;
  data.sample_size = buffer.whdr->bit_p_spl;
  data.stereo = (buffer.whdr->modus == 2);
  free(buffer.whdr);
  
  if ( (audio = open(AUDIODEV,O_WRONLY,0)) < 0 ) {
    err(1,"Error opening audio-device");
    return;
  }

  if ( ioctl(audio,SNDCTL_DSP_GETBLKSIZE,&abuf_size) < 0 ) {
    err(1,"Obtaining dsp buffer size.");
    goto errxit;
  }

  if ( abuf_size < 4096 || abuf_size > 65536 ) {
    sprintf(tbuf,"Invalid audio buffer size (%d bytes).",(int)abuf_size);
    err(1,tbuf);
    goto errxit;
  }

  if (!(iobuf = malloc(abuf_size))) {
    err(1,"Unable to allocate audio buffer.");
    goto errxit;
  }

  tmp = data.sample_size;
  if ( ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &data.sample_size) < 0 ) {
    err(1,"Unable to set audio sample size.");
    goto errxit;
  }

  if ( tmp != data.sample_size ) { /* Is this test needed? */
    err(0,"Unable to set sample size");
    goto errxit;
  }

  if ( ioctl(audio, SNDCTL_DSP_STEREO, &data.stereo) < 0 ) {
    err(1,"Unable to set sterio/mono.");
    goto errxit;
  }

  if ( ioctl(audio,SNDCTL_DSP_SPEED,&data.speed) < 0 ) {
    err(1,"Unable to set audio sampling rate.");
    goto errxit;
  }

  while (1) {
    /* Wait until buffer is not empty */
    if (buffer.readloop == ACTIVE_FINISH) {
      if (sem_trywait(&buffer.sem_read)) goto xit;
    } else
      sem_wait(&buffer.sem_read);
    /* Read the data and advance read pointer */
    write(audio,buffer.buffer[buffer.readpos], buffersize);
    free(buffer.buffer[buffer.readpos]);
    if (verbose) {putchar('f');fflush(stdout);}
    buffer.readpos++;
    if (buffer.readpos >= maxconbuffers) buffer.readpos = 0;
    /* Signal that the buffer has now one more location for writing */
    sem_post(&buffer.sem_write);
  }

xit:
  close(audio);
  return;

errxit:
  close(audio);
  exit(1);
}

void init(void)
{
  sem_init(&buffer.sem_write, 0, maxconbuffers);
  sem_init(&buffer.sem_read, 0, 0);
  buffer.readpos = 0;
  buffer.writepos = 0;
  buffer.readloop = ACTIVE_INIT;
}

void deinit(void)
{
  sem_destroy(&buffer.sem_write);
  sem_destroy(&buffer.sem_read);
}

int playmp3(char * filename)
{
  if (!quiet) print_mp3info(filename);
  init();  
  if (pthread_create(&read_thread,  0,
                    readloop,  filename, NULL)) {
    fputs("Cannot create thread\n",stderr);
    return 1;
  }
  sleep(writewait);
  writeloop(filename);
  deinit();
  return 0;
}

int main (int argc, char **argv)
{
  int c;

  binname = argv[0];
  signal(SIGPIPE, (void *)player_crash);

  /* Open config file*/
  {
    FILE * file;
    char cfgfile[256];
            
    strcpy(cfgfile, GLOBCFGFILE);
    if (file = fopen(cfgfile, "r"))
    {
      parsecfgfile(file);
      fclose(file);
    }

    strcpy(cfgfile, (char*)getenv("HOME"));
    strcat(cfgfile, LOCCFGFILE);
    if (file = fopen(cfgfile, "r"))
    {
      parsecfgfile(file);
      fclose(file);
    }
  }
                                                                          
  while (1)
    {
      int option_index = 0;
      static struct option long_options[] =
      {
        {"write-wait",    1, 0, 'w'},
        {"maxconbuffers", 1, 0, 'c'},
        {"buffersize",    1, 0, 's'},
        {"quiet",         0, 0, 'q'},
        {"help",          0, 0, 'h'},
        {"version",       0, 0, 'v'},
        {"verbose",       0, &verbose, 1},
        {0, 0, 0, 0}
      };

      c = getopt_long (argc, argv, "w:c:s:hv",
        long_options, &option_index);
      if (c == -1)
        break;

      switch (c)
        {
        case 'q': 
          quiet = 1; 
          break;

        case 'v':
          version();

	case 'w':
          if (sscanf(optarg,"%d",&writewait) != 1 || 
                writewait < MINWRITEWAIT)
            option_req_argument();
          break;

	case 'c': 
          if (sscanf(optarg,"%d",&maxconbuffers) != 1 || 
              maxconbuffers < MINMAXCONBUFFERS)
            option_req_argument();
          break;

	case 's': 
          if (sscanf(optarg,"%d",&buffersize) != 1 || 
              buffersize < MINBUFFERSIZE)
            option_req_argument();
          break;
        }
    }

    /* writewait */
    if(writewait < MINWRITEWAIT)
    {
      fprintf(stderr, "%s: invalid write wait specified, using default\n",binname);
      writewait = DEFWRITEWAIT;  /* Use default */
    }

    /* maxconbuffers */
    if(maxconbuffers < MINMAXCONBUFFERS)
    {
      fprintf(stderr, "%s: invalid maximum nuber of concurrent buffers specified, using default\n",binname);
      maxconbuffers = DEFMAXCONBUFFERS;  /* Use default */
    }

    /* buffersize */
    if(buffersize < MINBUFFERSIZE)
    {
      fprintf(stderr, "%s: invalid buffersize specified, using default\n",binname);
      buffersize = DEFBUFFERSIZE;  /* Use default */
    }
    
    if (optind == argc) usage(1);
    if (!(buffer.buffer = malloc(maxconbuffers*sizeof(int *)))) {
      err(1,"Unable to allocate pointer array");
    }
    buffersize*=1024;

    /* play 'em */
    while (optind < argc)
      playmp3(argv[optind++]);
    free(buffer.buffer);

  return EXIT_SUCCESS;
}

void parsecfgfile(FILE *File)
{
	char Buffer[256];	
	unsigned int i;
	char c;

	while(!feof(File))
	{
		/* Read string */		
		fscanf(File, "%s", Buffer);      		
	
		/* Comment */
		if(Buffer[0] == '#')
		{
			/* Got to next line */			
			fgets(Buffer, 255, File);
		} 

		/* writewait <sec> */
		else if(!strcasecmp(Buffer, "WriteWait"))
		{			
			fscanf(File, "%d ", &writewait);
		} 
		
		/* buffersize */
		else if(!strcasecmp(Buffer, "BufferSize"))
		{			
			fscanf(File, "%d ", &buffersize);
		} 
		
		/* maximum number of concurrent buffers */
		else if(!strcasecmp(Buffer, "MaxConBuffers"))
		{			
			fscanf(File, "%d ", &maxconbuffers);
		} 
		
		/* Unknown keyword */
		else
		{
			for(i=0, c=0; i < strlen(Buffer); i++)
				c += isgraph(Buffer[i]);
			if(c != 0)
			{
				fprintf(stderr, "%s: unknown keyword '%s'\n", binname,Buffer);
				exit(1);
			}
		}	      
	}
}

